@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////
 //
 // SFML - Simple and Fast Multimedia Library
-// Copyright (C) 2007-2016 Laurent Gomila (laurent@sfml-dev.org)
+// Copyright (C) 2007-2018 Laurent Gomila (laurent@sfml-dev.org)
 //
 // This software is provided 'as-is', without any express or implied warranty.
 // In no event will the authors be held liable for any damages arising from the use of this software.
@@ -61,6 +61,21 @@ namespace
     void close(FT_Stream)
     {
     }
+
+    // Helper to intepret memory as a specific type
+    template <typename T, typename U>
+    inline T reinterpret(const U& input)
+    {
+        T output;
+        std::memcpy(&output, &input, sizeof(U));
+        return output;
+    }
+
+    // Combine outline thickness, boldness and codepoint into a single 64-bit key
+    sf::Uint64 combine(float outlineThickness, bool bold, sf::Uint32 codePoint)
+    {
+        return (static_cast<sf::Uint64>(reinterpret<sf::Uint32>(outlineThickness)) << 32) | (static_cast<sf::Uint64>(bold) << 31) | codePoint;
+    }
 }
 
 
@@ -151,19 +166,21 @@ bool Font::loadFromFile(const std::string& filename)
     if (FT_Stroker_New(static_cast<FT_Library>(m_library), &stroker) != 0)
     {
         err() << "Failed to load font \"" << filename << "\" (failed to create the stroker)" << std::endl;
+        FT_Done_Face(face);
         return false;
     }
-    m_stroker = stroker;
 
     // Select the unicode character map
     if (FT_Select_Charmap(face, FT_ENCODING_UNICODE) != 0)
     {
         err() << "Failed to load font \"" << filename << "\" (failed to set the Unicode character set)" << std::endl;
+        FT_Stroker_Done(stroker);
         FT_Done_Face(face);
         return false;
     }
 
     // Store the loaded font in our ugly void* :)
+    m_stroker = stroker;
     m_face = face;
 
     // Store the font information
@@ -214,19 +231,21 @@ bool Font::loadFromMemory(const void* data, std::size_t sizeInBytes)
     if (FT_Stroker_New(static_cast<FT_Library>(m_library), &stroker) != 0)
     {
         err() << "Failed to load font from memory (failed to create the stroker)" << std::endl;
+        FT_Done_Face(face);
         return false;
     }
-    m_stroker = stroker;
 
     // Select the Unicode character map
     if (FT_Select_Charmap(face, FT_ENCODING_UNICODE) != 0)
     {
         err() << "Failed to load font from memory (failed to set the Unicode character set)" << std::endl;
+        FT_Stroker_Done(stroker);
         FT_Done_Face(face);
         return false;
     }
 
     // Store the loaded font in our ugly void* :)
+    m_stroker = stroker;
     m_face = face;
 
     // Store the font information
@@ -287,20 +306,23 @@ bool Font::loadFromStream(InputStream& stream)
     if (FT_Stroker_New(static_cast<FT_Library>(m_library), &stroker) != 0)
     {
         err() << "Failed to load font from stream (failed to create the stroker)" << std::endl;
+        FT_Done_Face(face);
+        delete rec;
         return false;
     }
-    m_stroker = stroker;
 
     // Select the Unicode character map
     if (FT_Select_Charmap(face, FT_ENCODING_UNICODE) != 0)
     {
         err() << "Failed to load font from stream (failed to set the Unicode character set)" << std::endl;
         FT_Done_Face(face);
+        FT_Stroker_Done(stroker);
         delete rec;
         return false;
     }
 
     // Store the loaded font in our ugly void* :)
+    m_stroker = stroker;
     m_face = face;
     m_streamRec = rec;
 
@@ -325,9 +347,7 @@ const Glyph& Font::getGlyph(Uint32 codePoint, unsigned int characterSize, bool b
     GlyphTable& glyphs = m_pages[characterSize].glyphs;
 
     // Build the key by combining the code point, bold flag, and outline thickness
-    Uint64 key = (static_cast<Uint64>(*reinterpret_cast<Uint32*>(&outlineThickness)) << 32)
-               | (static_cast<Uint64>(bold ? 1 : 0) << 31)
-               |  static_cast<Uint64>(codePoint);
+    Uint64 key = combine(outlineThickness, bold, codePoint);
 
     // Search the glyph into the cache
     GlyphTable::const_iterator it = glyphs.find(key);
@@ -504,7 +524,7 @@ void Font::cleanup()
     m_streamRec = NULL;
     m_refCount  = NULL;
     m_pages.clear();
-    m_pixelBuffer.clear();
+    std::vector<Uint8>().swap(m_pixelBuffer);
 }
 
 
@@ -551,7 +571,7 @@ Glyph Font::loadGlyph(Uint32 codePoint, unsigned int characterSize, bool bold, f
             FT_Stroker stroker = static_cast<FT_Stroker>(m_stroker);
 
             FT_Stroker_Set(stroker, static_cast<FT_Fixed>(outlineThickness * static_cast<float>(1 << 6)), FT_STROKER_LINECAP_ROUND, FT_STROKER_LINEJOIN_ROUND, 0);
-            FT_Glyph_Stroke(&glyphDesc, stroker, false);
+            FT_Glyph_Stroke(&glyphDesc, stroker, true);
         }
     }
 
@@ -583,11 +603,14 @@ Glyph Font::loadGlyph(Uint32 codePoint, unsigned int characterSize, bool bold, f
         // pollute them with pixels from neighbors
         const unsigned int padding = 1;
 
+        width += 2 * padding;
+        height += 2 * padding;
+
         // Get the glyphs page corresponding to the character size
         Page& page = m_pages[characterSize];
 
         // Find a good position for the new glyph into the texture
-        glyph.textureRect = findGlyphRect(page, width + 2 * padding, height + 2 * padding);
+        glyph.textureRect = findGlyphRect(page, width, height);
 
         // Make sure the texture data is positioned in the center
         // of the allocated texture rectangle
@@ -602,19 +625,32 @@ Glyph Font::loadGlyph(Uint32 codePoint, unsigned int characterSize, bool bold, f
         glyph.bounds.width  =  static_cast<float>(face->glyph->metrics.width)        / static_cast<float>(1 << 6) + outlineThickness * 2;
         glyph.bounds.height =  static_cast<float>(face->glyph->metrics.height)       / static_cast<float>(1 << 6) + outlineThickness * 2;
 
+        // Resize the pixel buffer to the new size and fill it with transparent white pixels
+        m_pixelBuffer.resize(width * height * 4);
+
+        Uint8* current = &m_pixelBuffer[0];
+        Uint8* end = current + width * height * 4;
+
+        while (current != end)
+        {
+            (*current++) = 255;
+            (*current++) = 255;
+            (*current++) = 255;
+            (*current++) = 0;
+        }
+
         // Extract the glyph's pixels from the bitmap
-        m_pixelBuffer.resize(width * height * 4, 255);
         const Uint8* pixels = bitmap.buffer;
         if (bitmap.pixel_mode == FT_PIXEL_MODE_MONO)
         {
             // Pixels are 1 bit monochrome values
-            for (int y = 0; y < height; ++y)
+            for (unsigned int y = padding; y < height - padding; ++y)
             {
-                for (int x = 0; x < width; ++x)
+                for (unsigned int x = padding; x < width - padding; ++x)
                 {
                     // The color channels remain white, just fill the alpha channel
-                    std::size_t index = (x + y * width) * 4 + 3;
-                    m_pixelBuffer[index] = ((pixels[x / 8]) & (1 << (7 - (x % 8)))) ? 255 : 0;
+                    std::size_t index = x + y * width;
+                    m_pixelBuffer[index * 4 + 3] = ((pixels[(x - padding) / 8]) & (1 << (7 - ((x - padding) % 8)))) ? 255 : 0;
                 }
                 pixels += bitmap.pitch;
             }
@@ -622,23 +658,23 @@ Glyph Font::loadGlyph(Uint32 codePoint, unsigned int characterSize, bool bold, f
         else
         {
             // Pixels are 8 bits gray levels
-            for (int y = 0; y < height; ++y)
+            for (unsigned int y = padding; y < height - padding; ++y)
             {
-                for (int x = 0; x < width; ++x)
+                for (unsigned int x = padding; x < width - padding; ++x)
                 {
                     // The color channels remain white, just fill the alpha channel
-                    std::size_t index = (x + y * width) * 4 + 3;
-                    m_pixelBuffer[index] = pixels[x];
+                    std::size_t index = x + y * width;
+                    m_pixelBuffer[index * 4 + 3] = pixels[x - padding];
                 }
                 pixels += bitmap.pitch;
             }
         }
 
         // Write the pixels to the texture
-        unsigned int x = glyph.textureRect.left;
-        unsigned int y = glyph.textureRect.top;
-        unsigned int w = glyph.textureRect.width;
-        unsigned int h = glyph.textureRect.height;
+        unsigned int x = glyph.textureRect.left - padding;
+        unsigned int y = glyph.textureRect.top - padding;
+        unsigned int w = glyph.textureRect.width + 2 * padding;
+        unsigned int h = glyph.textureRect.height + 2 * padding;
         page.texture.update(&m_pixelBuffer[0], w, h, x, y);
     }
 
@@ -689,10 +725,10 @@ IntRect Font::findGlyphRect(Page& page, unsigned int width, unsigned int height)
             if ((textureWidth * 2 <= Texture::getMaximumSize()) && (textureHeight * 2 <= Texture::getMaximumSize()))
             {
                 // Make the texture 2 times bigger
-                Image newImage;
-                newImage.create(textureWidth * 2, textureHeight * 2, Color(255, 255, 255, 0));
-                newImage.copy(page.texture.copyToImage(), 0, 0);
-                page.texture.loadFromImage(newImage);
+                Texture newTexture;
+                newTexture.create(textureWidth * 2, textureHeight * 2);
+                newTexture.update(page.texture);
+                page.texture.swap(newTexture);
             }
             else
             {
