@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////
 //
 // SFML - Simple and Fast Multimedia Library
-// Copyright (C) 2007-2017 Laurent Gomila (laurent@sfml-dev.org)
+// Copyright (C) 2007-2018 Laurent Gomila (laurent@sfml-dev.org)
 //
 // This software is provided 'as-is', without any express or implied warranty.
 // In no event will the authors be held liable for any damages arising from the use of this software.
@@ -88,7 +88,20 @@ m_hasMipmap    (false),
 m_cacheId      (getUniqueId())
 {
     if (copy.m_texture)
-        loadFromImage(copy.copyToImage());
+    {
+        if (create(copy.getSize().x, copy.getSize().y))
+        {
+            update(copy);
+
+            // Force an OpenGL flush, so that the texture will appear updated
+            // in all contexts immediately (solves problems in multi-threaded apps)
+            glCheck(glFlush());
+        }
+        else
+        {
+            err() << "Failed to copy texture, failed to create new texture" << std::endl;
+        }
+    }
 }
 
 
@@ -422,6 +435,115 @@ void Texture::update(const Uint8* pixels, unsigned int width, unsigned int heigh
 }
 
 
+////////////////////////////////////////////////////////////
+void Texture::update(const Texture& texture)
+{
+    // Update the whole texture
+    update(texture, 0, 0);
+}
+
+
+////////////////////////////////////////////////////////////
+void Texture::update(const Texture& texture, unsigned int x, unsigned int y)
+{
+    assert(x + texture.m_size.x <= m_size.x);
+    assert(y + texture.m_size.y <= m_size.y);
+
+    if (!m_texture || !texture.m_texture)
+        return;
+
+#ifndef SFML_OPENGL_ES
+
+    {
+        TransientContextLock lock;
+
+        // Make sure that extensions are initialized
+        priv::ensureExtensionsInit();
+    }
+
+    if (GLEXT_framebuffer_object && GLEXT_framebuffer_blit)
+    {
+        TransientContextLock lock;
+
+        // Save the current bindings so we can restore them after we are done
+        GLint readFramebuffer = 0;
+        GLint drawFramebuffer = 0;
+
+        glCheck(glGetIntegerv(GLEXT_GL_READ_FRAMEBUFFER_BINDING, &readFramebuffer));
+        glCheck(glGetIntegerv(GLEXT_GL_DRAW_FRAMEBUFFER_BINDING, &drawFramebuffer));
+
+        // Create the framebuffers
+        GLuint sourceFrameBuffer = 0;
+        GLuint destFrameBuffer = 0;
+        glCheck(GLEXT_glGenFramebuffers(1, &sourceFrameBuffer));
+        glCheck(GLEXT_glGenFramebuffers(1, &destFrameBuffer));
+
+        if (!sourceFrameBuffer || !destFrameBuffer)
+        {
+            err() << "Cannot copy texture, failed to create a frame buffer object" << std::endl;
+            return;
+        }
+
+        // Link the source texture to the source frame buffer
+        glCheck(GLEXT_glBindFramebuffer(GLEXT_GL_READ_FRAMEBUFFER, sourceFrameBuffer));
+        glCheck(GLEXT_glFramebufferTexture2D(GLEXT_GL_READ_FRAMEBUFFER, GLEXT_GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture.m_texture, 0));
+
+        // Link the destination texture to the destination frame buffer
+        glCheck(GLEXT_glBindFramebuffer(GLEXT_GL_DRAW_FRAMEBUFFER, destFrameBuffer));
+        glCheck(GLEXT_glFramebufferTexture2D(GLEXT_GL_DRAW_FRAMEBUFFER, GLEXT_GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, m_texture, 0));
+
+        // A final check, just to be sure...
+        GLenum sourceStatus;
+        glCheck(sourceStatus = GLEXT_glCheckFramebufferStatus(GLEXT_GL_READ_FRAMEBUFFER));
+
+        GLenum destStatus;
+        glCheck(destStatus = GLEXT_glCheckFramebufferStatus(GLEXT_GL_DRAW_FRAMEBUFFER));
+
+        if ((sourceStatus == GLEXT_GL_FRAMEBUFFER_COMPLETE) && (destStatus == GLEXT_GL_FRAMEBUFFER_COMPLETE))
+        {
+            // Blit the texture contents from the source to the destination texture
+            glCheck(GLEXT_glBlitFramebuffer(
+                0, texture.m_pixelsFlipped ? texture.m_size.y : 0, texture.m_size.x, texture.m_pixelsFlipped ? 0 : texture.m_size.y, // Source rectangle, flip y if source is flipped
+                x, y, x + texture.m_size.x, y + texture.m_size.y, // Destination rectangle
+                GL_COLOR_BUFFER_BIT, GL_NEAREST
+            ));
+        }
+        else
+        {
+            err() << "Cannot copy texture, failed to link texture to frame buffer" << std::endl;
+        }
+
+        // Restore previously bound framebuffers
+        glCheck(GLEXT_glBindFramebuffer(GLEXT_GL_READ_FRAMEBUFFER, readFramebuffer));
+        glCheck(GLEXT_glBindFramebuffer(GLEXT_GL_DRAW_FRAMEBUFFER, drawFramebuffer));
+
+        // Delete the framebuffers
+        glCheck(GLEXT_glDeleteFramebuffers(1, &sourceFrameBuffer));
+        glCheck(GLEXT_glDeleteFramebuffers(1, &destFrameBuffer));
+
+        // Make sure that the current texture binding will be preserved
+        priv::TextureSaver save;
+
+        // Set the parameters of this texture
+        glCheck(glBindTexture(GL_TEXTURE_2D, m_texture));
+        glCheck(glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_isSmooth ? GL_LINEAR : GL_NEAREST));
+        m_hasMipmap = false;
+        m_pixelsFlipped = false;
+        m_cacheId = getUniqueId();
+
+        // Force an OpenGL flush, so that the texture data will appear updated
+        // in all contexts immediately (solves problems in multi-threaded apps)
+        glCheck(glFlush());
+
+        return;
+    }
+
+#endif // SFML_OPENGL_ES
+
+    update(texture.copyToImage(), x, y);
+}
+
+
 ////////////////////////////////////////////////////////////
 void Texture::update(const Image& image)
 {
@@ -695,20 +817,30 @@ Texture& Texture::operator =(const Texture& right)
 {
     Texture temp(right);
 
-    std::swap(m_size,          temp.m_size);
-    std::swap(m_actualSize,    temp.m_actualSize);
-    std::swap(m_texture,       temp.m_texture);
-    std::swap(m_isSmooth,      temp.m_isSmooth);
-    std::swap(m_isRepeated,    temp.m_isRepeated);
-    std::swap(m_pixelsFlipped, temp.m_pixelsFlipped);
-    std::swap(m_fboAttachment, temp.m_fboAttachment);
-    std::swap(m_hasMipmap,     temp.m_hasMipmap);
-    m_cacheId = getUniqueId();
+    swap(temp);
 
     return *this;
 }
 
 
+////////////////////////////////////////////////////////////
+void Texture::swap(Texture& right)
+{
+    std::swap(m_size,          right.m_size);
+    std::swap(m_actualSize,    right.m_actualSize);
+    std::swap(m_texture,       right.m_texture);
+    std::swap(m_isSmooth,      right.m_isSmooth);
+    std::swap(m_sRgb,          right.m_sRgb);
+    std::swap(m_isRepeated,    right.m_isRepeated);
+    std::swap(m_pixelsFlipped, right.m_pixelsFlipped);
+    std::swap(m_fboAttachment, right.m_fboAttachment);
+    std::swap(m_hasMipmap,     right.m_hasMipmap);
+
+    m_cacheId = getUniqueId();
+    right.m_cacheId = getUniqueId();
+}
+
+
 ////////////////////////////////////////////////////////////
 unsigned int Texture::getNativeHandle() const
 {
